unit NLib;

{$IMPORTEDDATA OFF}
{$MINENUMSIZE 4}

interface

uses NTypes;

const
  Digits = ['0'..'9'];
  DigitsW = [WideChar('0')..WideChar('9')];
  WhiteSpaces = [#9..#13, ' '];
  WhiteSpacesW = [WideChar(#9)..WideChar(#13), WideChar(' ')];

  HexChars: array [0..15] of Char = '0123456789abcdef';

type
  { Exceptions }

  Exception = class(TObject)
    Msg: WideString;
  end;

  EError = class(Exception)
    constructor Create(const Msg: WideString);
  end;

  EValueError = class(Exception);

  ERuntimeError = class(Exception)
    constructor Create(ErrorCode: Byte; ErrorAddr: Pointer);
  end;

  { Types and Classes }

  TCmpInt = function (const A, B: Integer): Integer;
  TCmpStr = function (const A, B: string): Integer;
  TCmpWideStr = function (const A, B: WideString): Integer;

  TSeekWhence = (swSet, swCur, swEnd);

  TStream = class(TObject)
  protected
    ReadP, EndP, Buf: PChar;
  public
    destructor Destroy; override;
    function Read(var Buf; NBytes: Cardinal): Cardinal; virtual; abstract;
    function Write(const Buf; NBytes: Cardinal): Cardinal; virtual; abstract;
    function Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64; virtual;
        abstract;
    function Seekable: Boolean; virtual; abstract;
    function ReadLine(var S: string): Boolean;
  end;

  TBytesStream = class(TStream)
    Data: PChar;
    DataSize: Cardinal;
    Position: Cardinal;
    HasData: Boolean;
    constructor Create(const InitialBytes; NBytes: Cardinal;
        NoCopy: Boolean=False); overload;
    constructor Create(const S: string; NoCopy: Boolean=False); overload;
    destructor Destroy; override; 
    function Read(var Buf; NBytes: Cardinal): Cardinal; override;
    function ReadNoCopy(var Buf: PChar; NBytes: Cardinal): Cardinal;
    function Write(const Buf; NBytes: Cardinal): Cardinal; override;
    function Seek(Offset: Int64; Whence: TSeekWhence = swSet): Int64; override;
    function Seekable: Boolean; override;
  end;

  TArgParser = object
  private
    ArgInd: Integer;
    FArgs: TWideStringArray;
    Opt: PWideChar;
  public
    function Args: TWideStringArray;
    function GetArg: WideString;
    function GetOpt(var Opt_: WideChar): Boolean;
  end;

function ArgParser(): TArgParser; overload;
function ArgParser(const Args: array of WideString): TArgParser; overload;

procedure RaiseRuntimeError(ErrorCode: Byte; ErrorAddr: Pointer);

function BSwap16(X: Word): Word;
function BSwap32(X: Cardinal): Cardinal;
function BSwap64(X: Int64): Int64;

const
  Be16ToH: function (X: Word): Word = BSwap16;
  Be32ToH: function (X: Cardinal): Cardinal = BSwap32;
  Be64ToH: function (X: Int64): Int64 = BSwap64;
  HToBe16: function (X: Word): Word = BSwap16;
  HToBe32: function (X: Cardinal): Cardinal = BSwap32;
  HToBe64: function (X: Int64): Int64 = BSwap64;

function DivMod(A, B: Integer; var R: Integer): Integer;
function GCD(A, B: Cardinal): Cardinal;

function SplitCommandLine(const S: string): TStringArray;
function SplitCommandLineW(const S: WideString): TWideStringArray;
function JoinCommandLine(const Args: array of WideString): WideString;

function IStr(Value: Integer): string; overload;
function IStr(Value: Int64): string; overload;
function UIStr(Value: Cardinal): string;
function FStr(Value: Double; Digits: Integer): string;
function IHex(Value: Integer): string; overload;
function IHex(Value: Int64): string; overload;
function UIHex(Value: Cardinal): string;
function SHex(const S: string): string;
function UnitFormat(Value: Int64): string;
function NumFormat(Value: Int64): string;

function SInt(const S: string): Integer; overload;
function SInt(const S: WideString): Integer; overload;
function SUInt(const S: string): Cardinal; overload;
function SUInt(const S: WideString): Cardinal; overload;
function SInt64(const S: string): Int64; overload;
function SInt64(const S: WideString): Int64; overload;
function SFloat(const S: string): Double;
function HInt(const S: string): Integer;
function HBytes(const S: string): string;

function MemCmp(P1, P2: Pointer; NBytes: Cardinal): Integer;
function MemScan(P: Pointer; C: Integer; NBytes: Cardinal): Pointer;
function FindBytes(const Buf; NBytes: Integer; const Sub: string;
    Start: Integer=0): Integer;
function StartsWith(const S, Sub: WideString): Boolean;

function SliceArray(const StrArray: array of string; Start: Integer = 0):
    TStringArray;
function SliceArrayW(const StrArray: array of WideString; Start: Integer = 0):
    TWideStringArray;

function BaseName(const Path: WideString): WideString;
function DirName(const Path: WideString): WideString;

function Index(IntArray: array of Integer; Value: Integer): Integer; overload;
function Index(StrArray: array of string; const Value: String): Integer;
    overload;
function Cmp(const A, B: Integer): Integer; overload;
function Cmp(const A, B: string): Integer; overload;
function Cmp(const A, B: WideString): Integer; overload;
procedure Sort(var IntArray: array of Integer; Cmp: TCmpInt=nil); overload;
procedure Sort(var StrArray: array of string; Cmp: TCmpStr=nil); overload;
procedure Sort(var WideStrArray: array of WideString; Cmp: TCmpWideStr=nil); overload;

function FormatException(const E: Exception): WideString;
procedure SysFatal(Code: Integer = 1);
procedure SysWarn(Code: Integer = 1);

function Format(const S: string; const Args: array of const): string;
procedure RemoveLineSep(var S: string);
procedure RStrip(var S: string);

function IsEqualGUID(const A, B: TGUID): Boolean;

implementation

uses NSys;

const
  RuntimeErrorNames: array [0..24] of PChar = (
    'None',
    'Out of memory',
    'Invalid ptr',
    'Div by zero',
    'Range error',
    'Int overflow',
    'Invalid op',
    'Zero divide',
    'Overflow',
    'Underflow',
    'Invalid cast',
    'Access violation',
    'Priv instruction',
    'Control break',
    'Stack overflow',
    'Var type cast',
    'Var invalid op',
    'Var dispatch',
    'Var array create',
    'Var not array',
    'Var array bounds',
    'Assertion failed',
    'External exception',
    'Intf cast error',
    'Safe call error'
  );

procedure RaiseRuntimeError(ErrorCode: Byte; ErrorAddr: Pointer);
begin
  raise ERuntimeError.Create(ErrorCode, ErrorAddr);
end;

function BSwap16(X: Word): Word;
asm
  rol ax, 8
end;

function BSwap32(X: Cardinal): Cardinal;
asm
  bswap eax
end;

function BSwap64(X: Int64): Int64;
begin
  Result := Int64(BSwap32(X)) shl 32 or BSwap32(X shr 32);
end;

function DivMod(A, B: Integer; var R: Integer): Integer;
asm
  // Result := A div B;
  // R := A mod B;
  push ebx
  mov ebx, edx
  cdq
  idiv ebx
  mov [ecx], edx
  pop ebx
end;

function GCD(A, B: Cardinal): Cardinal;
begin
  Result := A;
  while B <> 0 do begin
    A := Result;
    Result := B;
    B := A mod B;
  end;
end;

function SplitCommandLine(const S: string): TStringArray;
var
  argc, bksc: Integer;
  p, ep: PChar;
  quoted: Boolean;
  t: string;

  procedure addbks(n: Integer);
  begin
    if n > 0 then
      t := t + StringOfChar('\', n);
  end;
begin
  argc := 0;
  bksc := 0;
  p := @S[1];
  ep := @S[Length(S) + 1];
  quoted := False;
  t := '';
  SetLength(Result, 1024);
  while p < ep do begin
    if (p^ in WhiteSpaces) and not quoted then begin
      addbks(bksc);
      if t <> '' then begin
        Result[argc] := t;
        Inc(argc);
        if argc = Length(Result) then
          SetLength(Result, argc + 1024);
        t := '';
      end;
    end else
      case p^ of
        '"':
          begin
            addbks(bksc shr 1);
            if (bksc and 1) = 1 then
              t := t + p^
            else if quoted and (p[1] = '"') then begin
              t := t + p^;
              Inc(p);
            end else
              quoted := not quoted;
            bksc := 0;
          end;
        '\': Inc(bksc);
      else
        addbks(bksc);
        bksc := 0;
        t := t + p^;
      end;
    Inc(p);
  end;
  addbks(bksc);
  if t <> '' then begin
    Result[argc] := t;
    Inc(argc);
  end;
  SetLength(Result, argc);
end;

function SplitCommandLineW(const S: WideString): TWideStringArray;
var
  argc, bksc: Integer;
  p, ep: PWideChar;
  quoted: Boolean;
  t: WideString;

  procedure addbks(n: Integer);
  begin
    if n > 0 then
      t := t + StringOfChar(WideChar('\'), n);
  end;
begin
  argc := 0;
  bksc := 0;
  p := @S[1];
  ep := @S[Length(S) + 1];
  quoted := False;
  t := '';
  SetLength(Result, 1024);
  while p < ep do begin
    if (p^ in WhiteSpacesW) and not quoted then begin
      addbks(bksc);
      bksc := 0;
      if t <> '' then begin
        Result[argc] := t;
        Inc(argc);
        if argc = Length(Result) then
          SetLength(Result, argc + 1024);
        t := '';
      end;
    end else
      case p^ of
        '"':
          begin
            addbks(bksc shr 1);
            if (bksc and 1) = 1 then
              t := t + p^
            else if quoted and (p[1] = '"') then begin
              t := t + p^;
              Inc(p);
            end else
              quoted := not quoted;
            bksc := 0;
          end;
        '\': Inc(bksc);
      else
        addbks(bksc);
        bksc := 0;
        t := t + p^;
      end;
    Inc(p);
  end;
  addbks(bksc);
  if t <> '' then begin
    Result[argc] := t;
    Inc(argc);
  end;
  SetLength(Result, argc);
end;

function JoinCommandLine(const Args: array of WideString): WideString;
  function quote(const S: WideString; QuoteCount: Integer): WideString;
  var
    i, j: Integer;
  begin
    SetString(Result, nil, Length(S) + QuoteCount);
    j := 1;
    for i := 1 to Length(S) do begin
      if S[i] = '"' then begin
        Result[j] := '"';
        Inc(j);
      end;
      Result[j] := S[i];
      Inc(j);
    end;
  end;
var
  i, j, len: Integer;
  space: Boolean;
  quotecount: Integer;
  quoted: TWideStringArray;
  p: PWideChar;
begin
  len := 0;
  SetLength(quoted, Length(Args));
  for i := 0 to High(Args) do begin
    space := False;
    quotecount := 0;
    for j := 1 to Length(Args[i]) do
      case Args[i][j] of
        #9, ' ':
          if (quotecount and 1) = 0 then
            space := True;
        '"': Inc(quotecount);
      end;
    if space or ((quotecount and 1) = 1) then
      quoted[i] := '"' + quote(Args[i], quotecount) + '"'
    else
      quoted[i] := Args[i];
    Inc(len, Length(quoted[i]));
  end;

  if len = 0 then begin
    Result := '';
    Exit;
  end;

  SetLength(Result, len + High(Args));
  p := PWideChar(Result);
  i := 0;
  while True do begin
    Move(quoted[i][1], p^, Length(quoted[i]) * SizeOf(WideChar));
    Inc(p, Length(quoted[i]));
    if i = High(quoted) then
      Break;
    p^ := ' ';
    Inc(p);
    Inc(i);
  end;
end;

function IStr(Value: Integer): string;
var
  buf: array [0..10] of Char;
asm
  {
  ep := @PChar(@buf)[SizeOf(buf)];
  p := ep;
  v := Value;
  Value := Abs(Value);
  repeat
    r := Value mod 10;
    Value := Value div 10;
    Dec(p);
    p^ := Char(r + Ord('0'));
  until Value = 0;
  if v < 0 then
  begin
    Dec(p);
    p^ := '-';
  end;
  SetString(Result, p, ep - p);
  }

  push ebx
  push edi
  mov ebx, eax
  lea edi, buf + 11 // end of buf + 1
  push edi
  push edx // Result
  mov ecx, 10
  // abs
  cdq // M = edx = (eax < 0) ? -1 : 0
  xor eax, edx // eax = x ^ M
  sub eax, edx // eax = x ^ M - M

@loop:
  xor edx, edx
  div ecx
  add edx, '0'
  dec edi
  mov byte ptr [edi], dl
  test eax, eax
  jnz @loop

  test ebx, ebx
  jge @exit
  dec edi
  mov byte ptr [edi], '-'

@exit:
  pop eax // Result
  mov edx, edi
  pop ecx // end of buf + 1
  sub ecx, edx
  call System.@LStrFromPCharLen
  pop edi
  pop ebx
end;

function IStr(Value: Int64): string;
var
  buf: array [0..19] of Char;
  p, ep: PChar;
  r: Integer;
  v: Int64;
begin
  ep := @PChar(@buf)[SizeOf(buf)];
  p := ep;
  v := Abs(Value);
  repeat
    r := v mod 10;
    v := v div 10;
    Dec(p);
    p^ := Char(r + Ord('0'));
  until v = 0;
  if Value < 0 then begin
    Dec(p);
    p^ := '-';
  end;
  SetString(Result, p, ep - p);
end;

function UIStr(Value: Cardinal): string;
var
  buf: array [0..10] of Char;
asm
  {
  p := ep;
  repeat
    r := Value mod 10;
    Value := Value div 10;
    Dec(p);
    p^ := Char(r + Ord('0'));
  until Value = 0;
  SetString(Result, p, ep - p);
  }

  push edi
  lea edi, buf + 11 // end of buf + 1
  push edi
  push edx // Result
  mov ecx, 10

@loop:
  xor edx, edx
  div ecx
  add edx, '0'
  dec edi
  mov byte ptr [edi], dl
  test eax, eax
  jnz @loop

  pop eax // Result
  mov edx, edi
  pop ecx // end of buf + 1
  sub ecx, edx
  call System.@LStrFromPCharLen
  pop edi
end;

function _Pow10(N: Integer): Double;
const
  tab: array [0..69] of Double = (
    1e0,  1e1,  1e2,  1e3,  1e4,  1e5,  1e6,  1e7,  1e8,  1e9,
    1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18, 1e19,
    1e20, 1e21, 1e22, 1e23, 1e24, 1e25, 1e26, 1e27, 1e28, 1e29,
    1e30, 1e31, 1e32, 1e33, 1e34, 1e35, 1e36, 1e37, 1e38, 1e39,
    1e40, 1e41, 1e42, 1e43, 1e44, 1e45, 1e46, 1e47, 1e48, 1e49,
    1e50, 1e51, 1e52, 1e53, 1e54, 1e55, 1e56, 1e57, 1e58, 1e59,
    1e60, 1e61, 1e62, 1e63, 1e64, 1e65, 1e66, 1e67, 1e68, 1e69
  );
var
  m: Integer;
begin
  if N < 0 then begin
    N := -N;
    if N < Length(tab) then begin
      Result := 1 / tab[N];
      Exit;
    end;
    m := n div 2;
    Result := 1 / (_Pow10(m) * _Pow10(N - m));
    Exit;
  end;
  if N < Length(tab) then begin
    Result := tab[N];
    Exit;
  end;
  m := n div 2;
  Result := _Pow10(m) * _Pow10(N - m);
end;

function FStr(Value: Double; Digits: Integer): string;
type
  TValue = record
    vl, vh: Cardinal; // little endian
  end;
  PValue = ^TValue;
const
  rlog2 = 0.30102999566398114; // 1/log2(10)
var
  m: Double;
  i, ndig, e2, e10, l, len: Integer;
  sp, p: PChar;
label
  exit_;
begin
  with PValue(@Value)^ do begin
    e2 := vh shr 20 and $7ff; // biased exponent
    if e2 = $7ff then begin
      if vl = 0 then
        if vh = $7ff00000 then begin
          Result := 'Inf';
          Exit;
        end else if vh = $fff00000 then begin
          Result := '-Inf';
          Exit;
        end;
      Result := 'NaN';
      Exit;
    end;
  end;

  {
    Value = 2^e2 * f = 10^e10 * m
    e10 = e2 * 1/log2(10)
    m = Value * 1/10^e10
  }
  Dec(e2, 1022); // unbiased
  e10 := Trunc(e2 * rlog2);
  m := Value * _Pow10(-e10);

  len := Digits + 3; // +3: '-' and '0.'
  if e10 > 0 then
    Inc(len, e10);
  SetString(Result, nil, len);
  sp := PChar(Result);
  p := sp;

  if m <> 0 then begin
    if m < 0 then begin
      m := -m;
      p^ := '-';
      Inc(p);
    end;

    while m < 1.0 do begin
      m := m * 10.0;
      Dec(e10);
    end;
    while m > 10.0 do begin
      m := m / 10.0;
      Inc(e10);
    end;

    m := m + _Pow10(-e10 - Digits - 1); // round off
  end;

  ndig := 0;
  if e10 < 0 then begin
    p^ := '0';
    Inc(p);
    if Digits = 0 then
      goto exit_;
    p^ := '.';
    Inc(p);

    for i := 1 to -e10 - 1 do begin
      if ndig = Digits then
        goto exit_;
      p^ := '0';
      Inc(p);
      Inc(ndig);
    end;
  end;

  l := 0;
  while True do begin
    i := Trunc(m);
    if not ((i = 0) and (l = 0) and (e10 > 0)) then begin
      p^ := Char(i + Ord('0'));
      Inc(p);
    end;
    if l > e10 then begin
      Inc(ndig);
      if ndig = Digits then
        Break;
    end;
    if l = e10 then begin
      if Digits = 0 then
        Break;
      p^ := '.';
      Inc(p);
    end;
    Inc(l);
    m := (m - i) * 10;
  end;

exit_:
  SetLength(Result, p - sp);
end;

function IHex(Value: Integer): string;
var
  buf: array [0..9] of Char;
asm
  {
  ep := @PChar(@buf)[SizeOf(buf)];
  p := ep;
  v := Value;
  Value := Abs(Value);
  i := SizeOf(buf);
  repeat
    Dec(i);
    buf[i] := HexChars[Value and $F];
    Value := Value shr 4;
  until Value = 0;
  if v < 0 then
  begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, p, ep - p);
  }

  push ebx
  push edi
  mov ebx, eax
  lea edi, buf + 10 // end of buf + 1
  push edi
  push edx // Result
  // abs
  cdq // M = edx = (eax < 0) ? -1 : 0
  xor eax, edx // eax = x ^ M
  sub eax, edx // eax = x ^ M - M

@loop:
  dec edi
  mov ecx, eax
  and ecx, $f
  mov cl, byte [HexChars + ecx]
  mov byte ptr [edi], cl
  shr eax, 4
  jnz @loop

  test ebx, ebx
  jge @exit
  dec edi
  mov byte ptr [edi], '-'

@exit:
  pop eax // Result
  mov edx, edi
  pop ecx // end of buf + 1
  sub ecx, edx
  call System.@LStrFromPCharLen
  pop edi
  pop ebx
end;

function IHex(Value: Int64): string;
var
  i: Integer;
  buf: array [0..17] of Char;
  v: Int64;
begin
  v := Abs(Value);
  i := SizeOf(buf);
  repeat
    Dec(i);
    buf[i] := HexChars[v and $F];
    v := v shr 4;
  until v = 0;
  if Value < 0 then begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, PChar(@buf[i]), SizeOf(buf)-i);
end;

function UIHex(Value: Cardinal): string;
var
  buf: array [0..9] of Char;
asm
  push edi
  lea edi, buf + 10 // end of buf + 1
  push edi
  push edx // Result

@loop:
  dec edi
  mov ecx, eax
  and ecx, $f
  mov cl, byte [HexChars + ecx]
  mov byte ptr [edi], cl
  shr eax, 4
  jnz @loop

  pop eax // Result
  mov edx, edi
  pop ecx // end of buf + 1
  sub ecx, edx
  call System.@LStrFromPCharLen
  pop edi
end;

function SHex(const S: string): string;
var
  i, len: Integer;
begin
  len := Length(S);
  SetString(Result, nil, len * 2);
  for i := 0 to len - 1 do begin
    Result[i*2+1] := HexChars[Byte(S[i+1]) shr 4];
    Result[i*2+2] := HexChars[Byte(S[i+1]) and $f];
  end;
end;

// convert to 4 digits or 3 digits + point and unit character
function UnitFormat(Value: Int64): string;
const
  ki = 1024;
  mi = 1024 * 1024;
  gi = 1024 * 1024 * 1024;
  pow: array [0..2] of Double = (1, 10, 100);
var
  u: Char;
  d: Integer;
  n: Double;
  v: Int64;
begin
  v := Abs(Value);

  if v < ki then begin
    Result := IStr(Value);
    Exit;
  end;

  if v >= mi * 1000 then begin
    n := v / gi;
    u := 'G';
  end else if v >= ki * 1000 then begin
    n := v / mi;
    u := 'M';
  end else begin
    n := v / ki;
    u := 'K';
  end;

  // truncate n
  if n >= 100 then
    d := 0
  else if n >= 10 then
    d := 1
  else
    d := 2;
  n := Int(n * pow[d]) / pow[d];

  if Value < 0 then
    n := -n;
  Result := FStr(n, d) + u;
end;

function NumFormat(Value: Int64): string;
const
  n: Integer = 3;
var
  s: string;
  i, len, offset: Integer;
  p: PChar;
begin
  s := IStr(Value);
  len := Length(s);
  SetLength(Result, len + (len-1) div n);
  p := PChar(Result);
  offset := n - len mod n;
  for i := 1 to len do begin
    p^ := s[i];
    if (i + offset) mod n = 0 then begin
      Inc(p);
      p^ := ',';
    end;
    Inc(p);
  end;
end;

function SInt(const S: string): Integer;
var
  sign: Boolean;
  i, j, len: Integer;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if i <= len then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in Digits) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
  if sign then
    Result := -Result;
end;

function SInt(const S: WideString): Integer;
var
  sign: Boolean;
  i, j, len: Integer;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if i <= len then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in DigitsW) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
  if sign then
    Result := -Result;
end;

function SUInt(const S: string): Cardinal;
var
  i, j, len: Integer;
begin
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) and (S[i] = '+') then
    Inc(i);
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in Digits) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
end;

function SUInt(const S: WideString): Cardinal;
var
  i, j, len: Integer;
begin
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if (i <= len) and (S[i] = '+') then
    Inc(i);
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in DigitsW) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
end;

function SInt64(const S: string): Int64;
var
  sign: Boolean;
  i, j, len: Integer;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if i <= len then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in Digits) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
  if sign then
    Result := -Result;
end;

function SInt64(const S: WideString): Int64;
var
  sign: Boolean;
  i, j, len: Integer;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if i <= len then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;
  j := i;
  Result := 0;
  while (i <= len) and (S[i] in DigitsW) do begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpacesW) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
  if sign then
    Result := -Result;
end;

function SFloat(const S: string): Double;
var
  sign: Boolean;
  i, j, k, len: Integer;
  b, d1, d2, f1, f2: Double;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if i <= len then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;

  k := i;

  d1 := 0;
  d2 := 0;
  j := 0;
  while (i <= len) and (S[i] in Digits) do begin
    if j < 8 then // 8: log10(2^53) / 2
      d1 := d1 * 10 + Ord(S[i]) - Ord('0')
    else begin
      d1 := d1 * 10;
      d2 := d2 * 10 + Ord(S[i]) - Ord('0');
    end;
    Inc(i);
    Inc(j);
  end;

  f1 := 0;
  f2 := 0;
  b := 1;
  if (i <= len) and (S[i] = '.') then begin
    Inc(i);
    j := 0;
    while (i <= len) and (S[i] in Digits) do begin
      if j < 8 then // 8: log10(2^53) / 2
        f1 := f1 * 10 + Ord(S[i]) - Ord('0')
      else begin
        f1 := f1 * 10;
        f2 := f2 * 10 + Ord(S[i]) - Ord('0');
      end;
      Inc(i);
      Inc(j);
      b := b * 10;
    end;
  end;

  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) or (k = i) then
    raise EValueError.Create;

  Result := d1 + d2 + (f1 + f2) / b;
  if sign then
    Result := -Result;
end;

function HInt(const S: string): Integer;
var
  sign: Boolean;
  i, j, len: Integer;
  d: Byte;
begin
  sign := False;
  i := 1;
  len := Length(S);
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if i <= Length(S) then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then begin
      sign := True;
      Inc(i);
    end;
  j := i;
  Result := 0;
  d := 0; // suppress warning
  while i <= Length(S) do begin
    case S[i] of
      '0'..'9': d := Ord('0');
      'A'..'F': d := Ord('A') - 10;
      'a'..'f': d := Ord('a') - 10;
    else
      Break;
    end;
    Result := Result * 16 + Ord(S[i]) - d;
    Inc(i);
  end;
  while (i <= len) and (S[i] in WhiteSpaces) do
    Inc(i);
  if (i <= len) or (j = i) then
    raise EValueError.Create;
  if sign then
    Result := -Result;
end;

function HBytes(const S: string): string;
var
  len: Integer;
  d: Byte;
  p, ep: PChar;
  q: PByte;
begin
  len := Length(S);
  if len and 1 <> 0 then
    raise EValueError.Create;
  SetLength(Result, len div 2);
  p := @S[1];
  ep := @p[len];
  q := @Result[1];
  while p < ep do begin
    case p^ of
      '0'..'9': d := Ord('0');
      'A'..'F': d := Ord('A') - 10;
      'a'..'f': d := Ord('a') - 10;
    else
      raise EValueError.Create;
    end;
    q^ := Ord(p^) - d;
    Inc(p);
    case p^ of
      '0'..'9': d := Ord('0');
      'A'..'F': d := Ord('A') - 10;
      'a'..'f': d := Ord('a') - 10;
    else
      raise EValueError.Create;
    end;
    q^ := q^ shl 4 or (Ord(p^) - d);
    Inc(p);
    Inc(q);
  end;
end;

function MemCmp(P1, P2: Pointer; NBytes: Cardinal): Integer;
asm
  push esi
  push edi
  mov edi, eax
  mov esi, edx
  xor eax, eax
  cld
  repe cmpsb
  je @exit
  sbb eax, eax // 0 if cf=0, -1 if cf=1
  xor eax, -2 // 0 xor -2 = -2, -1 xor -2 = 1
@exit:
  pop edi
  pop esi
end;

function MemScan(P: Pointer; C: Integer; NBytes: Cardinal): Pointer;
asm
  push edi
  mov edi, eax
  mov eax, edx
  test ecx, ecx
  jz @exit
  cld
  repne scasb
  jne @exit
  dec edi
@exit:
  mov eax, edi
  pop edi
end;

function FindBytes(const Buf; NBytes: Integer; const Sub: string;
    Start: Integer=0): Integer;
var
  i, j, endi, sublen: Integer;
  p: PChar;
begin
  Result := -1;
  sublen := Length(Sub);
  if sublen = 0 then
    Exit;
  endi := NBytes - sublen + 1;
  if endi < Start then
    Exit;
  p := @Buf;
  for i := Start to endi do
    if p[i] = Sub[1] then begin
      j := 1;
      while (j < sublen) and (p[i+j] = Sub[1+j]) do 
        Inc(j);
      if j = sublen then begin
        Result := i;
        Exit;
      end;
    end;
end;

function StartsWith(const S, Sub: WideString): Boolean;
begin
  Result := MemCmp(@S[1], @Sub[1], Length(Sub) * SizeOf(WideChar)) = 0;
end;

function SliceArray(const StrArray: array of string; Start: Integer = 0):
    TStringArray;
var
  i, len: Integer;
begin
  len := High(StrArray) - Start + 1;
  if len < 0 then
    len := 0;
  SetLength(Result, len);
  for i := 0 to High(Result) do
    Result[i] := StrArray[Start+i];
end;

function SliceArrayW(const StrArray: array of WideString; Start: Integer = 0):
    TWideStringArray;
var
  i, len: Integer;
begin
  len := High(StrArray) - Start + 1;
  if len < 0 then
    len := 0;
  SetLength(Result, len);
  for i := 0 to High(Result) do
    Result[i] := StrArray[Start+i];
end;

function BaseName(const Path: WideString): WideString;
var
  s: WideString;
begin
  SplitPath(Path, s, Result);
end;

function DirName(const Path: WideString): WideString;
var
  s: WideString;
begin
  SplitPath(Path, Result, s);
end;

function Index(IntArray: array of Integer; Value: Integer): Integer;
var
  i: Integer;
begin
  for i := Low(IntArray) to High(IntArray) do
    if IntArray[i] = Value then begin
      Result := i;
      Exit;
    end;
  Result := -1;
end;

function Index(StrArray: array of string; const Value: String): Integer;
var
  i: Integer;
begin
  for i := Low(StrArray) to High(StrArray) do
    if StrArray[i] = Value then begin
      Result := i;
      Exit;
    end;
  Result := -1;
end;

function Cmp(const A, B: Integer): Integer;
begin
  if A > B then
    Result := 1
  else if A < B then
    Result := -1
  else
    Result := 0;
end;

function Cmp(const A, B: string): Integer;
begin
  if A > B then
    Result := 1
  else if A < B then
    Result := -1
  else
    Result := 0;
end;

function Cmp(const A, B: WideString): Integer;
begin
  if A > B then
    Result := 1
  else if A < B then
    Result := -1
  else
    Result := 0;
end;

procedure Sort(var IntArray: array of Integer; Cmp: TCmpInt=nil);
  procedure merge(var Src, Dest: array of Integer; L, H: Integer);
  var
    i, j, k, m: Integer;
  begin
    if L >= H then
      Exit;
    m := (H + L) div 2;
    merge(Dest, Src, L, m);
    merge(Dest, Src, m+1, H);
    j := L;
    k := m+1;
    for i := L to H do begin
      if (k > H) or (j <= m) and (Cmp(Src[j], Src[k]) <= 0) then begin
        Dest[i] := Src[j];
        Inc(j);
      end else begin
        Dest[i] := Src[k];
        Inc(k);
      end;
    end;
  end;
var
  i: Integer;
  aux: array of Integer;
begin
  if not Assigned(Cmp) then
    Cmp := NLib.Cmp;
  SetLength(aux, Length(IntArray));
  for i := 0 to High(aux) do
    aux[i] := IntArray[i];
  merge(aux, IntArray, 0, High(aux));
end;

procedure Sort(var StrArray: array of string; Cmp: TCmpStr=nil);
  procedure merge(var Src, Dest: array of string; L, H: Integer);
  var
    i, j, k, m: Integer;
  begin
    if L >= H then
      Exit;
    m := (H + L) div 2;
    merge(Dest, Src, L, m);
    merge(Dest, Src, m+1, H);
    j := L;
    k := m+1;
    for i := L to H do begin
      if (k > H) or (j <= m) and (Cmp(Src[j], Src[k]) <= 0) then begin
        Dest[i] := Src[j];
        Inc(j);
      end else begin
        Dest[i] := Src[k];
        Inc(k);
      end;
    end;
  end;
var
  i: Integer;
  aux: array of string;
begin
  if not Assigned(Cmp) then
    Cmp := NLib.Cmp;
  SetLength(aux, Length(StrArray));
  for i := 0 to High(aux) do
    aux[i] := StrArray[i];
  merge(aux, StrArray, 0, High(aux));
end;

procedure Sort(var WideStrArray: array of WideString; Cmp: TCmpWideStr=nil);
  procedure merge(var Src, Dest: array of WideString; L, H: Integer);
  var
    i, j, k, m: Integer;
  begin
    if L >= H then
      Exit;
    m := (H + L) div 2;
    merge(Dest, Src, L, m);
    merge(Dest, Src, m+1, H);
    j := L;
    k := m+1;
    for i := L to H do begin
      if (k > H) or (j <= m) and (Cmp(Src[j], Src[k]) <= 0) then begin
        Dest[i] := Src[j];
        Inc(j);
      end else begin
        Dest[i] := Src[k];
        Inc(k);
      end;
    end;
  end;
var
  i: Integer;
  aux: array of WideString;
begin
  if not Assigned(Cmp) then
    Cmp := NLib.Cmp;
  SetLength(aux, Length(WideStrArray));
  for i := 0 to High(aux) do
    aux[i] := WideStrArray[i];
  merge(aux, WideStrArray, 0, High(aux));
end;

function FormatException(const E: Exception): WideString;
begin
  if E.Msg = '' then
    Result := E.ClassName + ' at ' + UIHex(Cardinal(ExceptAddr))
  else
    Result := E.Msg;
  if E is EOSError and (EOSError(E).Path <> '') then
    Result := EOSError(E).Path + ': ' + Result;
end;

procedure SysFatal(Code: Integer);
begin
  SysWarn(Code);
  Halt;
end;

procedure SysWarn(Code: Integer);
begin
  if ExceptObject is EPipeError then
    ExitCode := 128 + 13 // 128 + SIGPIPE
  else begin
    Print(StdErr, ProgName + ': ' + FormatException(Exception(ExceptObject)));
    ExitCode := Code;
  end;
end;

function Format(const S: string; const Args: array of const): string;
  function ToStr(const X: TVarRec; Typc: Char = 'd'): string;
  begin
    with X do
      case VType of
        vtAnsiString: Result := string(VAnsiString);
        vtChar: Result := VChar;
        vtInteger:
          case Typc of
            'x': Result := IHex(VInteger);
            'u': Result := UIStr(VInteger);
          else
            Result := IStr(VInteger);
          end;
        vtInt64:
          case Typc of
            'x': Result := IHex(VInt64^);
          else
            Result := IStr(VInt64^);
          end;
        vtWideString: Result := PWideChar(VWideString);
      end;
  end;
  function SIntDef(const S: string): Integer;
  begin
    try
      Result := SInt(S);
    except
      on EValueError do Result := 0;
    end;
  end;
type
  TState = (sIdx, sLit, sWid, sWidx, sTerm);
var
  i, autoidx, autowidxofs: Integer;
  state: TState;
  t, widstr, idxstr, padstr, widxstr: string;
  c, padc, typc, alignc: Char;
begin
  i := 0;
  autoidx := 0;
  autowidxofs := 0;
  state := sLit;
  typc := 'd';
  widstr := '';
  widxstr := '';
  idxstr := '';
  alignc := '<';
  Result := '';
  while i < Length(S) do begin
    Inc(i);
    c := S[i];
    case state of
      sLit:
        case c of
          '{':
            if S[i+1] = '{' then begin
              Result := Result + c;
              Inc(i);
            end else
              state := sIdx;
          '}':
            if S[i+1] = '}' then begin
              Result := Result + c;
              Inc(i);
            end;
        else
          Result := Result + c;
        end;
      sIdx:
        case c of
          '}': state := sTerm;
          ':': state := sWid;
          '0'..'9': idxstr := idxstr + c;
        end;
      sWid:
        case c of
          '{': state := sWidx;
          '}': state := sTerm;
          '0'..'9': widstr := widstr + c;
          '<', '>': alignc := c;
          'd', 'u', 'x': typc := c;
        end;
      sWidx:
        case c of
          '}':
            begin
              if widxstr = '' then begin
                if idxstr = '' then begin
                  Inc(autowidxofs);
                  widstr := widstr + ToStr(args[autoidx + autowidxofs]);
                end else begin
                  widstr := widstr + ToStr(args[autoidx]);
                  Inc(autoidx);
                end;
              end else
                widstr := widstr + ToStr(args[SIntDef(widxstr)]);
              widxstr := '';
              state := sWid;
            end;
          '0'..'9': widxstr := widxstr + c;
        end;
    end;
    if state = sTerm then begin
      if idxstr = '' then begin
        t := ToStr(args[autoidx], typc);
        Inc(autoidx);
      end else
        t := ToStr(args[SIntDef(idxstr)], typc);
      Inc(autoidx, autowidxofs);
      if (Length(widstr) > 0) and (widstr[1] = '0') then begin
        alignc := '>';
        padc := '0';
      end else
        padc := ' ';
      padstr := StringOfChar(padc, SIntDef(widstr)-Length(t));
      if alignc = '<' then
        Result := Result + t + padstr
      else
        Result := Result + padstr + t;
      idxstr := '';
      widstr := '';
      autowidxofs := 0;
      typc := 'd';
      alignc := '<';
      state := sLit;
    end;
  end;
end;

procedure RemoveLineSep(var S: string);
var
  len: Integer;
begin
  len := Length(S);
  if (len > 0) and (S[len] = #10) then begin
    Dec(len);
    if (len > 0) and (S[len] = #13) then
      Dec(len);
    SetLength(S, len);
  end;
end;

procedure RStrip(var S: string);
var
  i: Integer;
begin
  for i := Length(S) downto 1 do
    if not (S[i] in WhiteSpaces) then begin
      SetLength(S, i);
      Exit;
    end;
  S := '';
end;

function IsEqualGUID(const A, B: TGUID): Boolean;
begin
  Result := (A.D1 = B.D1) and
      (PCardinal(@A.D2)^ = PCardinal(@B.D2)^) and
      (PCardinal(@A.D4[0])^ = PCardinal(@B.D4[0])^) and
      (PCardinal(@A.D4[2])^ = PCardinal(@B.D4[2])^);
end;


{ Exceptions }

constructor EError.Create(const Msg: WideString);
begin
  Self.Msg := Msg;
end;

constructor ERuntimeError.Create(ErrorCode: Byte; ErrorAddr: Pointer);
var
  s: string;
begin
  if ErrorCode <= High(RuntimeErrorNames) then
    s := RuntimeErrorNames[ErrorCode]
  else
    s := 'Runtime error ' + UIStr(ErrorCode);
  Msg := s + ' at ' + UIHex(Cardinal(ErrorAddr));
end;


{ TStream }

destructor TStream.Destroy;
begin
  FreeMem(Buf);
  inherited;
end;

function TStream.ReadLine(var S: string): Boolean;
const
  bufsize = 8192;
var
  len, dlen: Integer;
  sp: PChar;
  foundlf: Boolean;
begin
  if Buf = nil then
    GetMem(Buf, bufsize);

  dlen := 0;
  foundlf := False;
  while not foundlf do begin
    if ReadP = EndP then begin
      len := Read(Buf^, bufsize);
      if len = 0 then begin
        Result := dlen <> 0;
        Exit;
      end;
      ReadP := Buf;
      EndP := Buf + len;
    end;

    sp := ReadP;
    ReadP := MemScan(sp, 10, EndP - sp);
    if ReadP < EndP then begin
      foundlf := True;
      Inc(ReadP);
    end;

    SetLength(S, dlen + (ReadP - sp));
    Move(sp^, S[dlen+1], ReadP - sp);
    Inc(dlen, ReadP - sp);
  end;
  Result := True;
end;


{ TBytesStream }

constructor TBytesStream.Create(const InitialBytes; NBytes: Cardinal;
    NoCopy: Boolean=False);
begin
  DataSize := NBytes;
  if NoCopy then
    Data := @InitialBytes
  else begin
    GetMem(Data, DataSize);
    HasData := True;
    Move(InitialBytes, Data^, DataSize);
  end;
  Seek(0);
end;

constructor TBytesStream.Create(const S: string; NoCopy: Boolean=False);
begin
  Create(S[1], Length(S), NoCopy);
end;

destructor TBytesStream.Destroy;
begin
  if HasData then
    FreeMem(Data);
end;

function TBytesStream.Read(var Buf; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Move(Data[Position], Buf, NBytes);
  Inc(Position, NBytes);
  Result := NBytes;
end;

function TBytesStream.ReadNoCopy(var Buf: PChar; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Buf := @Data[Position];
  Inc(Position, NBytes);
  Result := NBytes;
end;

function TBytesStream.Write(const Buf; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Move(Buf, Data[Position], NBytes);
  Inc(Position, NBytes);
  Result := NBytes;
end;

// TODO: support swCur and swEnd
function TBytesStream.Seek(Offset: Int64; Whence: TSeekWhence = swSet): Int64;
begin
  Position := Offset;
  if Position > DataSize then
    Position := DataSize;
  Result := Position;
end;

function TBytesStream.Seekable: Boolean;
begin
  Result := True;
end;

{ TArgParser }

function ArgParser: TArgParser;
begin
  Result.FArgs := Argv;
  Result.ArgInd := 1;
end;

function ArgParser(const Args: array of WideString): TArgParser;
var
  i: Integer;
begin
  SetLength(Result.FArgs, Length(Args));
  for i := 0 to High(Result.FArgs) do
    Result.FArgs[i] := Args[i];
end;

function TArgParser.Args: TWideStringArray;
var
  i, len: Integer;
begin
  len := Length(FArgs) - ArgInd;
  SetLength(Result, len);
  for i := 0 to len-1 do
    Result[i] := FArgs[ArgInd + i];
end;

function TArgParser.GetOpt(var Opt_: WideChar): Boolean;
begin
  if (Opt = nil) or (Opt^ = #0) then begin
    Result := False;

    if ArgInd = Length(FArgs) then
      Exit;

    Opt := PWideChar(FArgs[ArgInd]);
    if Opt^ <> '-' then
      Exit;

    Inc(Opt);
    if Opt^ = #0 then
      Exit;

    if (Opt^ = '-') and (Opt[1] = #0) then begin
      Inc(Opt);
      Inc(ArgInd);
      Exit;
    end;

    Inc(ArgInd);
  end;

  Opt_ := Opt^;
  Inc(Opt);
  Result := True;
end;

function TArgParser.GetArg: WideString;
begin
  if Opt^ = #0 then begin
    if ArgInd < Length(FArgs) then begin
      Result := FArgs[ArgInd];
      Inc(ArgInd);
    end else
      Result := '';
  end else
    Result := Opt;
  Opt := nil;
end;

initialization
  ErrorProc := @RaiseRuntimeError;
  ExceptObjProc := @_CreateSystemException;

finalization
  ErrorProc := nil;
  ExceptObjProc := nil;

end.
